import os
from osgeo import gdal, ogr
import numpy as np

def gdal_read_shapefile(sFilename_in, iFlag_metadata_only = 0):
    """
    Read a ESRI Shapefile. 
    Be careful that GDAL Python blinding has issue with some objects

    Args:
        sFilename_in (string): The filename

    Returns:
        tuple: aFeature, pSpatial_reference, lFeatureCount
    """

    if os.path.exists(sFilename_in):
        pass
    else:
        print('The file does not exist!')
        return

    sDriverName = "ESRI Shapefile"
    pDriver = ogr.GetDriverByName( sDriverName )

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName)

    pDataset = pDriver.Open(sFilename_in, gdal.GA_ReadOnly)
    
    # Check to see if shapefile is found.
    if pDataset is None:
        print ('Could not open %s' % (sFilename_in))
    else:
        pLayer = pDataset.GetLayer()
        lFeatureCount = pLayer.GetFeatureCount()
        print ("Number of features in %s: %d" %  (os.path.basename(sFilename_in),lFeatureCount))
        
        pSpatial_reference = pLayer.GetSpatialRef() 
        iFlag_first = 1

        for feature in pLayer:
            pGeometry = feature.GetGeometryRef()
            pEnvelope = pGeometry.GetEnvelope()

            if iFlag_first ==1:
                left_min = pEnvelope[0]
                right_max =  pEnvelope[1]
                bot_min =  pEnvelope[2]
                top_max =  pEnvelope[3]
                iFlag_first = 0
            else:
                left_min = np.min([left_min,  pEnvelope[0]])
                right_max = np.max([right_max,  pEnvelope[1]])
                bot_min = np.min([bot_min,  pEnvelope[2]])
                top_max = np.max([top_max,  pEnvelope[3]])

        if iFlag_metadata_only ==1:
            return pSpatial_reference, lFeatureCount, left_min, right_max, bot_min, top_max
        else:            
            aFeature=[]
            for pFeature in pLayer:         
                pGeometry = pFeature.GetGeometryRef()
                aFeature.append(pGeometry)       
                pass

            return aFeature, pSpatial_reference, lFeatureCount, left_min, right_max, bot_min, top_max


import os
import sys
from pyearth.system.define_global_variables import *


def slurm_prepare_job_script_python(iIndex_start,
                                    iIndex_end,
                                    sBasename_job,
                                    sDirectory_job,
                                    sFilename_python,
                                    iWalltime_in=None,
                                    nNode_in=None,
                                    nThread_in=None,
                                    sAccount=None,
                                    sEmail=None,
                                    sQueue_in=None):
    """
    Prepare a slurm job script for python without checkpoint

    Args:
        iIndex_start ([type]): The start index
        iIndex_end ([type]): The end index
        sBasename_job ([type]): The basename of job
        sDirectory_job ([type]): [description]
        sFilename_python ([type]): [description]
        iWalltime_in ([type], optional): [description]. Defaults to None.
        nNode_in ([type], optional): [description]. Defaults to None.
        nThread_in ([type], optional): [description]. Defaults to None.
        sAccount ([type], optional): [description]. Defaults to None.
        sEmail ([type], optional): [description]. Defaults to None.
        sQueue_in ([type], optional): [description]. Defaults to None.
    """
    if nNode_in is not None:
        iNode = nNode_in
    else:
        iNode = 1
    if nThread_in is not None:
        nThread = nThread_in
    else:
        nThread = 24

    if sQueue_in is not None:
        sQueue = sQueue_in
    else:
        sQueue = 'short'
    if iWalltime_in is not None:
        iWalltime = iWalltime_in
    else:
        iWalltime = 2

    nTask = iIndex_end - iIndex_start + 1
    nChunkPerThread = nTask // nThread

    sNode = "{:0d}".format(iNode)
    sTask = "{:0d}".format(nTask)
    sWalltime = "{:0d}".format(iWalltime)
    os.chdir(sDirectory_job)

    for iRank in range(1, nThread+1):
        if iRank == 1:
            iStart = (nThread-1) * nChunkPerThread + iIndex_start
            iEnd = iIndex_end
        else:
            iStart = (iRank-2) * nChunkPerThread + iIndex_start
            iEnd = (iRank-1) * nChunkPerThread + iIndex_start - 1
            sStart = "{:0d}".format(iStart)
            sEnd = "{:0d}".format(iEnd)

        sFilename_job = sDirectory_job + slash + \
            sBasename_job + "{:03d}".format(iRank) + '.job'
        pFile = open(sFilename_job, "w")  # write mode
        sLine = '#!/bin/bash' + '\n'
        pFile.write(sLine)

        if sAccount is not None:
            sLine = '#SBATCH --account=' + sAccount + '\n'
            pFile.write(sLine)

        sLine = '#SBATCH --begin=now+1minutes' + '\n'
        # pFile.write( sLine )

        sLine = '#SBATCH --cpus-per-task=1 ' + '\n'
        pFile.write(sLine)

        sLine = '#SBATCH --dependency=singleton ' + '\n'
        # pFile.write( sLine )
        sLine = '#SBATCH --error=stderr_%j.err' + '\n'
        pFile.write(sLine)
        sLine = '#SBATCH --job-name=' \
            + sBasename_job + \
                "{:03d}".format(iRank) + \
            '  # create a name for your job' + '\n'
        pFile.write(sLine)
        sLine = '#SBATCH --mail-type=ALL' + '\n'
        pFile.write(sLine)

        if sEmail is not None:
            sLine = '#SBATCH --mail-user=' + sEmail + '\n'
            pFile.write(sLine)

        sLine = '#SBATCH --nodes=' + sNode + ' # node count' + '\n'
        pFile.write(sLine)
        sLine = '#SBATCH --ntasks=1' + ' # total number of tasks' + '\n'
        pFile.write(sLine)
        sLine = '#SBATCH --output=stdout_%j.out' + '\n'
        pFile.write(sLine)

        sLine = '#SBATCH --partition=' + sQueue + '\n'  # can be improved here
        pFile.write(sLine)

        sLine = '#SBATCH --time=' + sWalltime + \
            ':00:00      # total run time limit (HH:MM:SS)' + '\n'
        pFile.write(sLine)

        sLine = 'module purge' + '\n'
        pFile.write(sLine)
        sLine = 'module load anaconda3/2019.03' + '\n'
        pFile.write(sLine)
        sLine = 'source /share/apps/anaconda3/2019.03/etc/profile.d/conda.sh' + '\n'
        pFile.write(sLine)
        sLine = 'unset PYTHONHOME' + '\n'
        pFile.write(sLine)
        sLine = 'conda activate gdalenv' + '\n'
        pFile.write(sLine)

        sLine = 'python ' + sFilename_python \
            + ' --iIndex_start ' + sStart \
            + ' --iIndex_end ' + sEnd + '\n'
        pFile.write(sLine)

        sLine = 'echo " Job " ' + '${SLURM_JOBID}' + ' is launched' + '\n'
        pFile.write(sLine)

        sLine = 'conda deactivate' + '\n'
        pFile.write(sLine)
        sLine = 'echo "Finished"' + '\n'
        pFile.write(sLine)
        pFile.close()

    return



import os
from osgeo import ogr
from pyearth.toolbox.management.vector.reproject import reproject_vector
from pyearth.toolbox.management.vector.merge_features import merge_features

def clip_vector_by_polygon_file(sFilename_vector_in, sFilename_polygon_in, sFilename_vector_out):
    # Open the input shapefile
    pDataset_source = ogr.Open(sFilename_vector_in)
    if pDataset_source is None:
        print("Error: Could not open the input shapefile.")
        return

    # Open the clip polygon
    pDataset_clip = ogr.Open(sFilename_polygon_in)
    if pDataset_clip is None:
        print("Error: Could not open the clip polygon.")
        return
    
     #get the extension of polygon file  
    sExtension_vector = os.path.splitext(sFilename_vector_in)[1]
    #get the driver for the extension
    if sExtension_vector == '.geojson':
        pDriver_vector = ogr.GetDriverByName('GeoJSON')
    else:
        pDriver_vector = ogr.GetDriverByName(sExtension_vector)
   
    #delete the output file if it exists
    if os.path.exists(sFilename_vector_out):
        pDriver_vector.DeleteDataSource(sFilename_vector_out)

    # Get the input layer and clip polygon layer
    pLayer_source = pDataset_source.GetLayer()    
    #get the spatial reference
    pSpatial_reference_target = pLayer_source.GetSpatialRef()    
    pProjection_target = pSpatial_reference_target.ExportToWkt()

    # Get the geometry of the clip polygon
    sExtension_clip = os.path.splitext(sFilename_polygon_in)[1]
    pLayer_clip = pDataset_clip.GetLayer()    
    pSpatial_reference_clip = pLayer_clip.GetSpatialRef()   
    #check the number the clip polygon
    nPolygon = pLayer_clip.GetFeatureCount()
    if nPolygon == 0:
        print('The polygon file does not contain any polygon!')
        return
    else:
        if nPolygon > 1:
            pDataset_clip = None
            pLayer_clip = None
            print('The polygon contains more than one polygon, the program will attempt to merge them as one!')
            #obtain the file extension
            sFilename_clip_new = sFilename_polygon_in.replace(sExtension_clip, '_merged' + sExtension_clip)
            merge_features(sFilename_polygon_in, sFilename_clip_new)
            sFilename_polygon_in = sFilename_clip_new
            #open the new file 
            pDataset_clip = ogr.Open(sFilename_polygon_in)                
            pLayer_clip = pDataset_clip.GetLayer(0) 
            pass

    # Get the spatial reference of the layer
    pSpatial_reference_clip = pLayer_clip.GetSpatialRef()
    pProjection_clip = pSpatial_reference_clip.ExportToWkt()
    
    if( pProjection_target != pProjection_clip):
        pDataset_clip = None
        pLayer_clip = None
        sFolder = os.path.dirname(sFilename_polygon_in)
        #get the name of the shapefile
        sName = os.path.basename(sFilename_polygon_in)
        #get the name of the shapefile without extension
        sName_no_extension = os.path.splitext(sName)[0]
        sFilename_clip_out = sFolder + '/' + sName_no_extension + '_transformed' + sExtension_vector
        reproject_vector(sFilename_polygon_in, sFilename_clip_out, pProjection_target)                
        pDataset_clip = ogr.Open(sFilename_clip_out)
        pLayer_clip = pDataset_clip.GetLayer(0) 
    else:
        
        pass
       

    pFeature_clip = pLayer_clip.GetNextFeature()
    pPolygon_clip = pFeature_clip.GetGeometryRef() 

    # Create a new output shapefile   
    pDataset_clipped = pDriver_vector.CreateDataSource(sFilename_vector_out)
    if pDataset_clipped is None:
        print("Error: Could not create the output shapefile.")
        return

    #get the layer
    pLayer_in = pDataset_source.GetLayer()
    #obtain the geotype 
    iGeomType = pLayer_in.GetGeomType()
    if iGeomType == ogr.wkbPoint:
        #create the layer
        pLayer_clipped = pDataset_clipped.CreateLayer('layer', pSpatial_reference_target, geom_type=ogr.wkbPoint)
    else:
        if iGeomType == ogr.wkbLineString:
            #create the layer
            pLayer_clipped = pDataset_clipped.CreateLayer('layer', pSpatial_reference_target, geom_type=ogr.wkbLineString)
        else:
            if iGeomType == ogr.wkbPolygon:
                #create the layer
                pLayer_clipped = pDataset_clipped.CreateLayer('layer', pSpatial_reference_target, geom_type=ogr.wkbPolygon)
            else:
                print('Geometry type not supported')
                return
            pass
        pass
  

    # Apply the clipping operation to each pFeature in the input shapefile
    for pFeature in pLayer_source:
        # Get the geometry of the input pFeature
        pGeometry_source = pFeature.GetGeometryRef()

        # Perform the clipping operation
        pGeometry_clip = pGeometry_source.Intersection(pPolygon_clip)
        # Create a new pFeature in the output layer with the clipped geometry
        if pGeometry_clip is not None and not pGeometry_clip.IsEmpty():
            #we only want to keep the clipped geometry that is within the clip polygon            
            pFeature_clip = ogr.Feature(pLayer_clipped.GetLayerDefn())
            pFeature_clip.SetGeometry(pGeometry_clip)
            pLayer_clipped.CreateFeature(pFeature_clip)
            pFeature_clip = None            
        else:
            #check pFeature that is entirely within the clip polygon
            if pPolygon_clip.Contains(pGeometry_source):
                pFeature_clip = ogr.Feature(pLayer_clipped.GetLayerDefn())
                pFeature_clip.SetGeometry(pGeometry_source)
                pLayer_clipped.CreateFeature(pFeature_clip)
                pFeature_clip = None        

    # Close the shapefiles
    pDataset_source = None
    pDataset_clip = None
    pDataset_clipped = None
    pSpatial_reference_clip = None
    pSpatial_reference_target = None

    print("Clipping completed.")





import numpy as np
def dt2cal(dt):
    """
    Convert array of datetime64 to a calendar array of year, month, day, hour,
    minute, seconds, microsecond with these quantites indexed on the last axis.

    Parameters
    ----------
    dt : datetime64 array (...)
        numpy.ndarray of datetimes of arbitrary shape

    Returns
    -------
    cal : uint32 array (..., 7)
        calendar array with last axis representing year, month, day, hour,
        minute, second, microsecond
    """

    # allocate output 
    out = np.empty(dt.shape + (7,), dtype="u4")
    # decompose calendar floors
    Y, M, D, h, m, s = [dt.astype(f"M8[{x}]") for x in "YMDhms"]
    out[..., 0] = Y + 1970 # Gregorian Year
    out[..., 1] = (M - Y) + 1 # month
    out[..., 2] = (D - M) + 1 # dat
    out[..., 3] = (dt - D).astype("m8[h]") # hour
    out[..., 4] = (dt - h).astype("m8[m]") # minute
    out[..., 5] = (dt - m).astype("m8[s]") # second
    out[..., 6] = (dt - s).astype("m8[us]") # microsecond
    return out



import math
from datetime import datetime

def __to_format(jd: float, fmt: str) -> float:
    """
    Converts a Julian Day object into a specific format.  For
    example, Modified Julian Day.
    Parameters
    ----------
    jd: float
    fmt: str

    Returns
    -------
    jd: float
    """
    if fmt.lower() == 'jd':
        return jd
    elif fmt.lower() == 'mjd':
        return jd - 2400000.5
    elif fmt.lower() == 'rjd':
        return jd - 2400000
    else:
        raise ValueError('Invalid Format')


def __from_format(jd: float, fmt: str):
    """
    Converts a Julian Day format into the "standard" Julian
    day format.
    Parameters
    ----------
    jd
    fmt

    Returns
    -------
    (jd, fractional): (int, float)
         A tuple representing a Julian day.  The first number is the
         Julian Day Number, and the second is the fractional component of the
         day.  A fractional component of 0.5 represents noon.  Therefore
         the standard julian day would be (jd + fractional + 0.5)
    """
    if fmt.lower() == 'jd':
        # If jd has a fractional component of 0, then we are 12 hours into
        # the day
        return math.floor(jd + 0.5), jd + 0.5 - math.floor(jd + 0.5)
    elif fmt.lower() == 'mjd':
        return __from_format(jd + 2400000.5, 'jd')
    elif fmt.lower() == 'rjd':
        return __from_format(jd + 2400000, 'jd')
    else:
        raise ValueError('Invalid Format')


def to_jd(dt: datetime, fmt: str = 'jd') -> float:
    """
    Converts a given datetime object to Julian date.
    Algorithm is copied from https://en.wikipedia.org/wiki/Julian_day
    All variable names are consistent with the notation on the wiki page.

    Parameters
    ----------
    fmt
    dt: datetime
        Datetime object to convert to MJD

    Returns
    -------
    jd: float
    """
    a = math.floor((14-dt.month)/12)
    y = dt.year + 4800 - a
    m = dt.month + 12*a - 3

    jdn = dt.day + math.floor((153*m + 2)/5) + 365*y + math.floor(y/4) - math.floor(y/100) + math.floor(y/400) - 32045

    jd = jdn + (dt.hour - 12) / 24 + dt.minute / 1440 + dt.second / 86400 + dt.microsecond / 86400000000

    return __to_format(jd, fmt)


def from_jd(jd: float, fmt: str = 'jd') -> datetime:
    """
    Converts a Julian Date to a datetime object.
    Algorithm is from Fliegel and van Flandern (1968)

    Parameters
    ----------
    jd: float
        Julian Date as type specified in the string fmt

    fmt: str

    Returns
    -------
    dt: datetime

    """
    jd, jdf = __from_format(jd, fmt)

    l = jd+68569
    n = 4*l//146097
    l = l-(146097*n+3)//4
    i = 4000*(l+1)//1461001
    l = l-1461*i//4+31
    j = 80*l//2447
    k = l-2447*j//80
    l = j//11
    j = j+2-12*l
    i = 100*(n-49)+i+l

    year = int(i)
    month = int(j)
    day = int(k)

    # in microseconds
    frac_component = int(jdf * (1e6*24*3600))

    hours = int(frac_component // (1e6*3600))
    frac_component -= hours * 1e6*3600

    minutes = int(frac_component // (1e6*60))
    frac_component -= minutes * 1e6*60

    seconds = int(frac_component // 1e6)
    frac_component -= seconds*1e6

    frac_component = int(frac_component)

    dt = datetime(year=year, month=month, day=day,
                  hour=hours, minute=minutes, second=seconds, microsecond=frac_component)
    return dt


import os
import sys

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from pyearth.system.define_global_variables import *
from pyearth.visual.color.create_diverge_rgb_color_hex import create_diverge_rgb_color_hex

def plot_time_series_analysis(aTime,
                              aData,
                              sFilename_out,
                              sVariable,
                              iDPI_in=None,
                              iFlag_without_raw_in=None,
                              iFlag_log_in=None,
                              iFlag_scientific_notation_in=None,
                              iReverse_y_in=None,
                              iSize_x_in=None,
                              iSize_y_in=None,
                              dMax_x_in=None,
                              dMin_x_in=None,
                              dMax_y_in=None,
                              dMin_y_in=None,
                              dSpace_x_in=None,
                              dSpace_y_in=None,
                              aMarker_in=None,
                              aColor_in=None,
                              aLinestyle_in=None,
                              sLabel_x_in=None,
                              sLabel_y_in=None,
                              aLabel_legend_in=None,
                              aLocation_legend_in=None,
                              sLocation_legend_in=None,
                              sDate_type_in=None,
                              sFormat_y_in=None,
                              sTitle_in=None):

    try:
        import pandas as pd
        from statsmodels.tsa.seasonal import STL
        from statsmodels.tsa.stattools import adfuller
    except ImportError as e:
        raise ImportError(
            "The package 'pandas and statsmodels' is required for this function to run.") from e

    aTime = np.array(aTime)
    aData = np.array(aData)
    pShape = aData.shape

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_without_raw_in is not None:
        iFlag_without_raw = iFlag_without_raw_in
    else:
        iFlag_without_raw = 0

    if iReverse_y_in is not None:
        iReverse_y = iReverse_y_in
    else:
        iReverse_y = 0

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if sLabel_y_in is not None:
        sLabel_y = sLabel_y_in
    else:
        sLabel_y = ''

    nData = 4
    if aLabel_legend_in is not None:
        aLabel_legend = aLabel_legend_in
    else:
        aLabel_legend = np.full(nData, '')

    if sTitle_in is not None:
        sTitle = sTitle_in.capitalize()  # title will also convert units
    else:
        sTitle = ''

    if aMarker_in is not None:
        aMarker = aMarker_in
    else:
        aMarker = ['o', '.', '*', '+']

    if aColor_in is not None:
        aColor = aColor_in
    else:
        aColor = create_diverge_rgb_color_hex(nData)

    if aLinestyle_in is not None:
        aLinestyle = aLinestyle_in
    else:
        aLinestyle = ['-', '--', '-.', 'solid']

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = np.datetime64(np.nanmax(aTime), 'Y')

    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = np.datetime64(np.nanmin(aTime), 'Y')

    if dMax_y_in is not None:
        dMax_y = dMax_y_in
    else:
        dMax_y = np.nanmax(aData)

    if dMin_y_in is not None:
        dMin_y = dMin_y_in
    else:
        dMin_y = np.nanmin(aData)

    if (dMax_y <= dMin_y):
        return
    else:
        dMin_y = dMin_y - 0.13 * (dMax_y-dMin_y)
        dMax_y = dMax_y + 0.13 * (dMax_y-dMin_y)

    if dSpace_y_in is not None:
        iFlag_space_y = 1
        dSpace_y = dSpace_y_in
    else:
        iFlag_space_y = 0
        dSpace_y = (dMax_y - dMin_y) / 4.0
        if dSpace_y < 1:
            pass
        else:
            dSpace_y = int(dSpace_y)
        pass

    adf_test = adfuller(aData)
    print("ADF = " + str(adf_test[0]))
    print("p-value = " + str(adf_test[1]))
    print("Critical value = ", (adf_test[4]))

    pYear = mpl.dates.YearLocator(1)   # every year
    pMonth = mpl.dates.MonthLocator()  # every month
    if sDate_type_in is not None:
        if sDate_type_in == 'month':
            pMonth = mpl.dates.MonthLocator(3)
        else:
            print(sDate_type_in)
    else:
        pass

    if sFormat_y_in is not None:
        iFlag_format_y = 1
        sFormat_y = sFormat_y_in
    else:
        iFlag_format_y = 0
        sFormat_y = '{:.1f}'

    if sLocation_legend_in is not None:
        sLocation_legend = sLocation_legend_in
    else:
        sLocation_legend = "upper right"

    if aLocation_legend_in is not None:
        aLocation_legend = aLocation_legend_in
    else:
        aLocation_legend = (1.0, 1.0)

    sYear_format = mpl.dates.DateFormatter('%Y')

    aTS = pd.date_range(aTime[0],         periods=len(aTime),      freq='M')
    aData_tsa = pd.Series(aData,    index=aTS,      name=sVariable)

    # https://www.statsmodels.org/stable/generated/statsmodels.tsa.seasonal.STL.html#statsmodels.tsa.seasonal.STL
    stl = STL(aData_tsa, seasonal=13)
    aTSA = stl.fit()
    # part 1
    # plot time series
    if iFlag_without_raw == 1:
        aLabel_y = np.array(['Trend', 'Season', 'Residual'])
        aData_all = [aTSA.trend, aTSA.seasonal, aTSA.resid]
        iSize_y = iSize_y * 0.75
        fig, pAxGrid = plt.subplots(nrows=len(aData_all),
                                           figsize=(iSize_x, iSize_y),
                                           sharex=True,  dpi=iDPI)

        for i, ax in enumerate(pAxGrid):
            tsp, = ax.plot(aTime, aData_all[i],
                           color=aColor[i+1], linestyle=aLinestyle[i+1],
                           marker=aMarker[i+1],
                           zorder=3)
            if i == 0:
                ax.set_title(sTitle, fontsize=13)

                # we will not use legend method because it has 3 panels
                if aLabel_legend_in is not None:
                    # plot the first on the top
                    sText = aLabel_legend_in[0]
                    dLocation = 0.95
                    ax.text(0.03, dLocation, sText,
                            verticalalignment='top', horizontalalignment='left',
                            transform=ax.transAxes,
                            color='black', fontsize=10)
                    # plot the remaining on the bot
                    nlegend = len(aLabel_legend_in)
                    for i in range(1, nlegend, 1):
                        sText = aLabel_legend_in[i]
                        dLocation = nlegend * 0.06 - i * 0.05 - 0.03
                        ax.text(0.03, dLocation, sText,
                                verticalalignment='top', horizontalalignment='left',
                                transform=ax.transAxes,
                                color='black', fontsize=10)
                        pass

            if i == 2:
                ax.plot((dMin_x, dMax_x), (0, 0),
                        color='#000000', linestyle=':', zorder=2)
                ax.set_xlabel('Year', fontsize=12)
            ax.set_ylabel(aLabel_y[i], fontsize=12)
            ax.grid(which='major', color='lightgrey',
                    linestyle=':', axis='y', zorder=1)
            ax.set_xlim(dMin_x, dMax_x)
            ax.xaxis.set_major_locator(pYear)
            ax.xaxis.set_minor_locator(pMonth)
            ax.xaxis.set_major_formatter(sYear_format)

    else:
        # include raw data
        aLabel_y = np.array([sLabel_y, 'Trend', 'Season', 'Residual'])
        aData_all = [aData, aTSA.trend, aTSA.seasonal, aTSA.resid]
        fig, pAxGrid = plt.subplots(nrows=len(aData_all),
                                           figsize=(iSize_x, iSize_y),
                                           sharex=True,  dpi=iDPI)
        for i, ax in enumerate(pAxGrid):
            tsp, = ax.plot(aTime, aData_all[i],
                           color=aColor[i], linestyle=aLinestyle[i],
                           marker=aMarker[i],
                           zorder=3)
            # the first plot has title
            if i == 0:
                # might have to consider log label here, refer to time series plot
                aTickLabel_y = list()
                ax.set_title(sTitle, fontsize=13)
                if iReverse_y == 1:
                    ax.set_ylim(dMin_y, dMax_y)

            # the bottom plot has x label and provided y label
            if i == 3:
                ax.plot((dMin_x, dMax_x), (0, 0),
                        color='#000000', linestyle=':', zorder=2)
                ax.set_xlabel('Year', fontsize=12)

            ax.set_ylabel(aLabel_y[i], fontsize=12)
            ax.grid(which='major', color='lightgrey',
                    linestyle=':', axis='y', zorder=1)
            ax.set_xlim(dMin_x, dMax_x)
            ax.xaxis.set_major_locator(pYear)
            ax.xaxis.set_minor_locator(pMonth)
            ax.xaxis.set_major_formatter(sYear_format)

    plt.savefig(sFilename_out, bbox_inches='tight')
    plt.close('all')
    plt.clf()



#it is also recommended to use the nvector API to calculate the area of a polygon on the ellipsoid
import numpy as np
from pyearth.system.define_global_variables import *

from pyearth.gis.geometry.calculate_spherical_triangle_area import calculate_spherical_triangle_area


def calculate_polygon_area(aLongitude_in,
                           aLatitude_in,
                           iFlag_algorithm=1,
                           iFlag_radian=None,
                           dRadius_in=None):
    """
    Computes area of spherical polygon, assuming spherical Earth. 
    Returns result in ratio of the sphere's area if the radius is specified. Otherwise, in the units of provided radius.
    lats and lons are in degrees.

    Args:
        aLongitude_in (list): The longitude of list of points
        aLatitude_in (list): The latitude of list of points
        iFlag_algorithm (int, optional): Which algorithm to use. Defaults to 0.
        radius (float, optional): The radius of Earth in meter. Defaults to 6378137.0.

    Returns:
        float: The area of the polygon
    """

    npoint = len(aLongitude_in)
    if npoint < 3:
        print('More than 2 points are required!')
        return

    # TODO: take into account geodesy (i.e. convert latitude to authalic sphere, use radius of authalic sphere instead of mean radius of spherical earth)
    # close polygon
    if aLatitude_in[0] != aLatitude_in[-1] or aLongitude_in[0] != aLongitude_in[-1]:
        aLatitude_in = np.append(aLatitude_in, aLatitude_in[0])
        aLongitude_in = np.append(aLongitude_in, aLongitude_in[0])

    if iFlag_radian is None:  # degree_based
        aLongitude_radian_in = np.deg2rad(aLongitude_in)
        aLatitude_radian_in = np.deg2rad(aLatitude_in)

    else:
        aLongitude_radian_in = aLongitude_in
        aLatitude_radian_in = aLatitude_in

        pass

    if iFlag_algorithm == 0:
        # Line integral based on Green's Theorem, assumes spherical Earth

        # Get colatitude (a measure of surface distance as an angle)
        a = np.sin(aLatitude_radian_in/2)**2 + np.cos(aLatitude_radian_in) * \
            np.sin(aLongitude_radian_in/2)**2
        colat = 2*np.arctan2(np.sqrt(a), np.sqrt(1-a))

        # azimuth of each point in segment from the arbitrary origin
        az = np.arctan2(np.cos(aLatitude_radian_in) * np.sin(aLongitude_in),
                        np.sin(aLatitude_radian_in)) % (2*np.pi)

        # Calculate step sizes
        # daz = np.diff(az) % (2*pi)
        daz = np.diff(az)
        daz = (daz + np.pi) % (2 * np.pi) - np.pi

        # Determine average surface distance for each step
        deltas = np.diff(colat)/2
        colat = colat[0:-1]+deltas

        # Integral over azimuth is 1-cos(colatitudes)
        integrands = (1-np.cos(colat)) * daz

        # Integrate and save the answer as a fraction of the unit sphere.
        # Note that the sum of the integrands will include a factor of 4pi.
        area = abs(sum(integrands))  # Could be area of inside or outside

        area = min(area, 1-area)

    elif iFlag_algorithm == 1:
        # L'Huilier Theorem, assumes spherical earth
        # see:
        # https://mathworld.wolfram.com/SphericalPolygon.html
        # https://web.archive.org/web/20160324191929/http://forum.worldwindcentral.com/showthread.php?20724-A-method-to-compute-the-area-of-a-spherical-polygon
        # https://github.com/spacetelescope/spherical_geometry/blob/master/spherical_geometry/polygon.py
        # https://github.com/tylerjereddy/spherical-SA-docker-demo/blob/master/docker_build/demonstration.py
        #

        dLongtitude_root = aLongitude_radian_in[0]
        dLatitude_root = aLatitude_radian_in[0]
        dLongtitude_b = aLongitude_radian_in[1]
        dLatitude_b = aLatitude_radian_in[1]
        nTriangle = npoint - 2
        aArea = np.zeros(nTriangle)
        for i in np.arange(1, nTriangle+1, 1):
            # define a triangle
            dLongtitude_a = dLongtitude_b
            dLatitude_a = dLatitude_b
            dLongtitude_b = aLongitude_radian_in[i+1]
            dLatitude_b = aLatitude_radian_in[i+1]
            # calculate the area of the triangle
            aLongitude_temp = [dLongtitude_root, dLongtitude_a, dLongtitude_b]
            aLatitude_temp = [dLatitude_root, dLatitude_a, dLatitude_b]
            dArea_triangle = calculate_spherical_triangle_area(aLongitude_temp,
                                                               aLatitude_temp,
                                                               iFlag_radian=1)

            aArea[i-1] = dArea_triangle

            pass

        area = np.sum(aArea)
        pass

    elif iFlag_algorithm == 2:
        # https://trs.jpl.nasa.gov/handle/2014/41271
        # TODO
        pass

    if iFlag_radian is not None:
        return area
    else:
        #6371229.0 or something else
        if dRadius_in is not None:
            dArea_m = area * dRadius_in**2
        else:
            dArea_m = area * earth_radius**2
        return dArea_m




if __name__ == '__main__':
    # test the polygon area calculation
    #aLongitude_in = [-148, -148, -148.5, -148.5]
    #aLatitude_in = [68, 68.5, 68.5, 68]
#
    #dArea = calculate_polygon_area(aLongitude_in, aLatitude_in, iFlag_algorithm=0)
    #print(dArea)
#
    #dArea1 = calculate_polygon_area(aLongitude_in, aLatitude_in)
    #print(dArea1)
#
    #print(dArea/dArea1)

    xc = -148.1875
    yc = 70.125
    dResolution = 1.0/16

    aLongitude_in = [xc-dResolution/2, xc+dResolution/2, xc+dResolution/2, xc-dResolution/2]
    aLatitude_in = [yc-dResolution/2, yc-dResolution/2, yc+dResolution/2, yc+dResolution/2]
    dArea1 = calculate_polygon_area(aLongitude_in, aLatitude_in)
    print(dArea1)




def world2Pixel(pGeoMatrix_in, dx_in, dy_in):
    """
    Uses a gdal geomatrix (gdal.GetGeoTransform()) to calculate
    the pixel location of a geospatial coordinate
   
    Args:
        geoMatrix (gdal): The geotransform matrix
        dx_in ([type]): The X coodinate of a point
        dy_in ([type]): The Y coodinate of a point

    Returns:
        Tuple: pixel, line
    """
    
    ulX = pGeoMatrix_in[0]
    ulY = pGeoMatrix_in[3]
    xDist = pGeoMatrix_in[1]
    yDist = pGeoMatrix_in[5]
    rtnX = pGeoMatrix_in[2]
    rtnY = pGeoMatrix_in[4]
    pixel = int((dx_in - ulX) / xDist)
    line = int((ulY - dy_in) / xDist)
    return (pixel, line)


import numpy as np
def degree_to_meter(dLatitude_mean_in, dResolution_degree_in):
    """[summary]

    Args:
        dLatitude_in (float): The latitide of interest
        dResolution_degree_in (float): The desired resolution in degree

    Returns:
        float: The resolution in meter at this latitude
    """
    dRadius = 6378137.0 
    dRadius2 = dRadius * np.cos( dLatitude_mean_in / 180.0 * np.pi)
    dResolution_meter = dResolution_degree_in / 360.0 * (2*np.pi * dRadius2)

    return dResolution_meter


def meter_to_degree(dLatitude_mean_in, dResolution_meter_in ):
    """[summary]

    Args:
        dResolution_meter_in (float): [description]
        dLatitude_mean_in (float): [description]

    Returns:
        float: The desired resoluton in degree
    """
    dLatitude_mean_in = abs(dLatitude_mean_in)

    dRadius = 6378137.0
    dRadius2 = dRadius * np.cos( dLatitude_mean_in / 180.0 * np.pi)

 
    dResolution_degree= dResolution_meter_in/(2*np.pi * dRadius2) * 360.0

    return dResolution_degree



import io
import os

from setuptools import setup, find_packages

NAME = "pyearth"
DESCRIPTION = \
    "Python for Earth Science."
AUTHOR = "Chang Liao"
AUTHOR_EMAIL = "changliao.climate@gmail.com"
URL = "https://github.com/changliao1025/pyearth"
VERSION = "0.1.25"
REQUIRES_PYTHON = ">=3.8.0"
KEYWORDS = "Earth Science"

REQUIRED = [    
    "numpy", 
    "gdal",    
    "matplotlib",
    "cartopy", 
]

CLASSIFY = [
    "Development Status :: 4 - Beta",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Visualization"
]

HERE = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(
            HERE, "README.md"), encoding="utf-8") as f:
        LONG_DESCRIPTION = "\n" + f.read()

except FileNotFoundError:
    LONG_DESCRIPTION = DESCRIPTION


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="custom",
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    python_requires=REQUIRES_PYTHON,
    keywords=KEYWORDS,
    url=URL,
    packages=find_packages(),
    install_requires=REQUIRED,
    classifiers=CLASSIFY,
    extras_require={
        'statistics': ['requests','netCDF4','pandas', 'scipy', 'statsmodels']
    }
)




import numpy as np

from pyearth.gis.location.convert_longitude_latitude_to_sphere_3d import convert_longitude_latitude_to_sphere_3d
from pyearth.gis.geometry.calculate_angle_between_vectors_degrees import calculate_angle_between_vectors_degrees

def calculate_angle_betwen_vertex(dLongitude1_in, dLatitude1_in, 
                                dLongitude2_in, dLatitude2_in, 
                                dLongitude3_in, dLatitude3_in, 
                                iFlag_radian = None):
    #all in degree
    if iFlag_radian is None:
        dLongitude1_radian_in, dLatitude1_radian_in = np.radians(np.array((dLongitude1_in, dLatitude1_in) ))
        dLongitude2_radian_in, dLatitude2_radian_in = np.radians(np.array((dLongitude2_in, dLatitude2_in) )) #this is the middle one
        dLongitude3_radian_in, dLatitude3_radian_in = np.radians(np.array((dLongitude3_in, dLatitude3_in) ))
        pass
    else:         
        dLongitude1_radian_in, dLatitude1_radian_in = dLongitude1_in , dLatitude1_in  
        dLongitude2_radian_in, dLatitude2_radian_in = dLongitude2_in , dLatitude2_in #this is the middle one
        dLongitude3_radian_in, dLatitude3_radian_in = dLongitude3_in , dLatitude3_in
        pass
    # The points in 3D space
    a3 = convert_longitude_latitude_to_sphere_3d(dLongitude1_radian_in, dLatitude1_radian_in)
    b3 = convert_longitude_latitude_to_sphere_3d(dLongitude2_radian_in, dLatitude2_radian_in)
    c3 = convert_longitude_latitude_to_sphere_3d(dLongitude3_radian_in, dLatitude3_radian_in)
    # Vectors in 3D space
    a3vec = a3 - b3
    c3vec = c3 - b3        
    angle3deg = calculate_angle_between_vectors_degrees(a3vec, c3vec)
    return  angle3deg



if __name__ == '__main__':

    dLongitude1_in=-148
    dLatitude1_in= 71
 
    dLongitude2_in=-148.1875
    dLatitude2_in=70.125
    
    dLongitude3_in=-147
    dLatitude3_in= 70.125
                          
    

    dAngle = calculate_angle_betwen_vertex(dLongitude1_in, dLatitude1_in, 
                                dLongitude2_in, dLatitude2_in, 
                                dLongitude3_in, dLatitude3_in 
                               )
    print(dAngle)


import numpy as np
from datetime import datetime

import matplotlib as mpl
import matplotlib.pyplot as plt
from pyearth.system.define_global_variables import *
from pyearth.visual.calculate_ticks_space import calculate_ticks_space

def plot_time_series_data_monthly_fill(aTime,
                                       aData,
                                       sFilename_out,
                                       iDPI_in=None,
                                       iFlag_trend_in=None,
                                       iReverse_Y_in=None,
                                       iSize_X_in=None,
                                       iSize_Y_in=None,
                                       dMax_Y_in=None,
                                       dMin_Y_in=None,
                                       sMarker_in=None,
                                       sLabel_Y_in=None,
                                       sLabel_legend_in=None,
                                       sTitle_in=None):

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_trend_in is not None:
        iFlag_trend = 1
    else:
        iFlag_trend = 0

    if iReverse_Y_in is not None:
        iReverse_Y = 1
    else:
        iReverse_Y = 0

    if iSize_X_in is not None:
        iSize_X = iSize_X_in
    else:
        iSize_X = 12
    if iSize_Y_in is not None:
        iSize_Y = iSize_Y_in
    else:
        iSize_Y = 9

    if sLabel_Y_in is not None:
        sLabel_Y = sLabel_Y_in
    else:
        sLabel_Y = ''
    if sLabel_legend_in is not None:
        sLabel_legend = sLabel_legend_in
    else:
        sLabel_legend = ''
    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    if sMarker_in is not None:
        sMarker = sMarker_in
    else:
        sMarker = '+'

    nstress = len(aTime)
    nan_index = np.where(aData == missing_value)
    aData[nan_index] = np.nan
    good_index = np.where(~np.isnan(aData))

    if dMax_Y_in is not None:
        dMax_Y = dMax_Y_in
    else:
        dMax_Y = np.nanmax(aData) * 1.2
    if dMin_Y_in is not None:
        dMin_Y = dMin_Y_in
    else:
        dMin_Y = np.nanmin(aData)  # if it has negative value, change here
    if (dMax_Y <= dMin_Y):
        return

    fig = plt.figure(dpi=iDPI)
    fig.set_figwidth(iSize_X)
    fig.set_figheight(iSize_Y)
    ax = fig.add_axes([0.1, 0.5, 0.8, 0.4])
    pYear = mpl.dates.YearLocator(5)   # every year
    pMonth = mpl.dates.MonthLocator()  # every month
    sYear_format = mpl.dates.DateFormatter('%Y')
    x1 = aTime
    y1 = (aData[1])[0]
    y_top = (aData[0])[0]
    y_bot = (aData[2])[0]
    ax.fill_between(x1, y_top, y_bot,  facecolor='cornflowerblue')
    ax.plot(x1, y1,
            color='red', linestyle='--',
            marker=sMarker, markeredgecolor='blue',
            label=sLabel_legend)

    # calculate linear regression
    if iFlag_trend == 1:
        x_dummy = np.array([i.timestamp() for i in x1])
        x_dummy = x_dummy[good_index]
        y_dummy = y1[good_index]
        coef = np.polyfit(x_dummy, y_dummy, 1)
        poly1d_fn = np.poly1d(coef)
        mn = np.min(x_dummy)
        mx = np.max(x_dummy)
        x2 = [mn, mx]
        y2 = poly1d_fn(x2)
        x2 = [datetime.fromtimestamp(i) for i in x2]
        ax.plot(x2, y2, color='orange', linestyle='-.',  linewidth=0.5)

    ax.axis('on')
    ax.grid(which='major', color='grey', linestyle='--', axis='y')
    # ax.grid(which='minor', color='#CCCCCC', linestyle=':') #only y axis grid is

    # ax.set_aspect(dRatio)  #this one set the y / x ratio
    ax.xaxis.set_major_locator(pYear)
    ax.xaxis.set_minor_locator(pMonth)
    ax.xaxis.set_major_formatter(sYear_format)
    ax.tick_params(axis="x", labelsize=10)
    ax.tick_params(axis="y", labelsize=10)

    ax.set_xmargin(0.05)
    ax.set_ymargin(0.15)

    ax.set_xlabel('Year', fontsize=12)
    ax.set_ylabel(sLabel_Y, fontsize=12)
    ax.set_title(sTitle, loc='center', fontsize=15)
    # round to nearest years...
    x_min = np.datetime64(aTime[0], 'Y')
    x_max = np.datetime64(aTime[nstress-1], 'Y') + np.timedelta64(1, 'Y')
    ax.set_xlim(x_min, x_max)
    if dMax_Y < 1000 and dMax_Y > 0.1:
        ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
    else:
        ax.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1e'))
    dummy = calculate_ticks_space(
        [y1, y_top, y_bot], nstep_in=5, iFlag_small_in=1)
    dSpace = dummy[0]
    if (dSpace <= 0):
        ax.invert_yaxis()

    else:
        ax.yaxis.set_major_locator(mpl.ticker.MultipleLocator(dSpace))
        dMin_Y = dummy[1]
        dMax_Y = dummy[2]
        if (iReverse_Y == 1):
            ax.set_ylim(dMax_Y, dMin_Y)
        else:
            ax.set_ylim(dMin_Y, dMax_Y)
    ax.legend(bbox_to_anchor=(1.0, 1.0), loc="upper right", fontsize=12)
    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    plt.clf()
    # print('finished plotting')



import os, sys
import numpy as np
from osgeo import gdal, ogr
from pyearth.toolbox.management.vector.merge_features import merge_features
from pyearth.toolbox.management.vector.reproject import reproject_vector
from pyearth.gis.gdal.read.raster.gdal_read_geotiff_file import gdal_read_geotiff_file


def clip_raster_by_polygon_file(sFilename_raster_in, sFilename_polygon_in, sFilename_raster_out, sFormat_in='GTiff'):
    """
    Clip a raster by a polygon
    :param sFilename_raster_in: input raster filename
    :param sFilename_polygon_in: input polygon filename
    :param sFilename_raster_out: output raster filename
    :param sFormat: output format
    :return: None
    """

    #check input files
    if os.path.exists(sFilename_raster_in):
        pass
    else:
        print('The raster file does not exist!')
        return
    
    if os.path.exists(sFilename_polygon_in):
        pass
    else:   
        print('The polygon file does not exist!')
        return
    
    #get the extension of raster file
    if sFormat_in is not None:
        sDriverName = sFormat_in
    else: 
        sDriverName = 'GTiff'
        sExtension_raster = os.path.splitext(sFilename_raster_in)[1]
    #get the driver for the extension    
    pDriver_raster = gdal.GetDriverByName(sDriverName)

    #get the extension of polygon file  
    sExtension_vector = os.path.splitext(sFilename_polygon_in)[1]
    #get the driver for the extension
    if sExtension_vector == '.geojson':
        pDriver_vector = ogr.GetDriverByName('GeoJSON')
    else:
        pDriver_vector = ogr.GetDriverByName(sExtension_vector)
   

    pDataset_data = gdal.Open(sFilename_raster_in, gdal.GA_ReadOnly)   
    dummy = gdal_read_geotiff_file(sFilename_raster_in)
    aData= dummy['dataOut']    
    eType = dummy['dataType']  
    dPixelWidth = dummy['pixelWidth']                        
    pPixelHeight = dummy['pixelHeight']
    dOriginX = dummy['originX']
    dOriginY = dummy['originY']
    nrow = dummy['nrow']
    ncolumn = dummy['ncolumn']
    dMissing_value= dummy['missingValue']
    pProjection = dummy['projection']
    pSpatial_reference_target = dummy['spatialReference']
    pProjection_target = pSpatial_reference_target.ExportToWkt()
    dX_left=dOriginX
    dX_right = dOriginX + ncolumn * dPixelWidth
    dY_top = dOriginY
    dY_bot = dOriginY + nrow * pPixelHeight

    #get the spatial reference of the shapefile
    pDataset_clip = ogr.Open(sFilename_polygon_in)    
    # Get the first layer in the shapefile
    pLayer_clip = pDataset_clip.GetLayer(0)   
    # Count the number of features (polygons)
    nPolygon = pLayer_clip.GetFeatureCount()
    if nPolygon == 0:
        print('The shapefile does not contain any polygon!')
        return
    else:
        if nPolygon > 1:
            pDataset_clip = None
            pLayer_clip = None
            print('The polygon contains more than one polygon, the program will attempt to merge them as one!')
            #obtain the file extension
            sFilename_clip_new = sFilename_polygon_in.replace(sExtension_vector, '_merged' + sExtension_vector)
            merge_features(sFilename_polygon_in, sFilename_clip_new)
            sFilename_polygon_in = sFilename_clip_new
            #open the new file 
            pDataset_clip = ogr.Open(sFilename_polygon_in)    
            # Get the first layer in the shapefile
            pLayer_clip = pDataset_clip.GetLayer(0) 
            pass

    # Get the spatial reference of the layer
    pSpatial_reference_clip = pLayer_clip.GetSpatialRef()
    pProjection_clip = pSpatial_reference_clip.ExportToWkt()
    print(pProjection_clip)   

    if( pProjection_target != pProjection_clip):        
        pDataset_clip = None
        pLayer_clip = None
        #in this case, we can reproject the shapefile to the same spatial reference as the raster
        #get the folder that contains the shapefile
        sFolder = os.path.dirname(sFilename_polygon_in)
        #get the name of the shapefile
        sName = os.path.basename(sFilename_polygon_in)
        #get the name of the shapefile without extension
        sName_no_extension = os.path.splitext(sName)[0]
        #create a new shapefile
        sFilename_clip_out = sFolder + '/' + sName_no_extension + '_transformed' + sExtension_vector
        reproject_vector(sFilename_polygon_in, sFilename_clip_out, pProjection_target)        
        #use the new shapefile to clip the raster        
        sFilename_clip = sFilename_clip_out
        pDataset_clip = ogr.Open(sFilename_clip_out)
        pLayer_clip = pDataset_clip.GetLayer(0) 
        pFeature_clip = pLayer_clip.GetNextFeature()
        pPolygon = pFeature_clip.GetGeometryRef() 

    else:
        sFilename_clip = sFilename_polygon_in              
        #read the first polygon 
        pFeature_clip = pLayer_clip.GetNextFeature()
        pPolygon = pFeature_clip.GetGeometryRef() 
        #get the envelope of the polygon
        iFlag_transform = 0

    #use the gdal warp function to clip the raster
    minX, maxX, minY, maxY = pPolygon.GetEnvelope()
    iNewWidth = int( (maxX - minX) / abs(dPixelWidth)  )
    iNewHeigh = int( (maxY - minY) / abs(dPixelWidth) )
    newGeoTransform = (minX, dPixelWidth, 0,  maxY, 0, -dPixelWidth)  
    if minX > dX_right or maxX < dX_left or minY > dY_top or maxY < dY_bot:        
        #this polygon is out of bound            
        pass
    else:       
        pDataset_clip = pDriver_raster.Create(sFilename_raster_out, iNewWidth, iNewHeigh, 1, eType)
        pDataset_clip.SetGeoTransform( newGeoTransform )
        pDataset_clip.SetProjection( pProjection)   
        pWrapOption = gdal.WarpOptions( cropToCutline=True,cutlineDSName = sFilename_clip , 
                width=iNewWidth,   
                    height=iNewHeigh,      
                        dstSRS=pSpatial_reference_target , format = 'MEM' )
        pDataset_clip_warped = gdal.Warp('', pDataset_data, options=pWrapOption)

        #convert the warped dataset to an array
        aData_clip = pDataset_clip_warped.ReadAsArray()
        #change the missing value to the original missing value
        aData_clip[aData_clip == dMissing_value] = -9999
        # Write the warped dataset to the output raster
        pDataset_clip.GetRasterBand(1).WriteArray(aData_clip)        
        #close the dataset
        pDataset_clip = None
        #close the dataset
        pDataset_data = None
        print('The raster has been clipped by the polygon file!')

    pSpatial_reference_target = None
    pSpatial_reference_clip = None
  
    

      

    


def check_if_duplicates(aList_in):
    """
    Check if given list contains any duplicates
    Args:
        aList_in (list): List of native Python data type

    Returns:
        int: 1-no duplicate, 0-has duplicate
    """

    iFlag_unique = 1
    for elem in aList_in:
        if aList_in.count(elem) > 1:
            iFlag_unique = 0
            break
        else:
            pass

    return iFlag_unique




import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from pyearth.toolbox.math.stat.scipy_bivariate_kde import scipy_bivariate_kde

def fmt0(x):
        a, b = '{:.1e}'.format(x).split('e')
        b = int(b)
        return r'${} \times 10^{{{}}}$'.format(a, b)

def fmt1(x):
        a = '{:.1f}'.format(x)
        return a

def fmt2(x):
        a = '{:.1e}'.format(x)
        return a

def scatter_plot_data_density(aData_x, 
                              aData_y,
                              sFilename_out, 
                              iSize_x_in = None, 
                              iSize_y_in = None, 
                              iDPI_in = None,
                              iFlag_scientific_notation_x_in=None,
                              iFlag_scientific_notation_y_in=None,
                              iFlag_log_x_in=None,
                              iFlag_log_y_in=None,
                              dMin_x_in = None, 
                              dMax_x_in = None, 
                              dMin_y_in = None, 
                              dMax_y_in = None, 
                              dSpace_x_in = None, 
                              dSpace_y_in = None, 
                              sFormat_x_in =None,
                              sFormat_y_in =None,
                              sLabel_x_in = None, 
                              sLabel_y_in = None, 
                              sLabel_legend_in = None, 
                              sTitle_in = None):
    try:
        import scipy
    except ImportError as e:
        raise ImportError("The package 'scipy' is required for this function to run.") from e

    if iSize_x_in is not None:
        iSize_x = iSize_x_in
    else:
        iSize_x = 12

    if iSize_y_in is not None:
        iSize_y = iSize_y_in
    else:
        iSize_y = 9

    if iDPI_in is not None:
        iDPI = iDPI_in
    else:
        iDPI = 300

    if iFlag_log_x_in is not None:
        iFlag_log_x = iFlag_log_x_in
    else:
        iFlag_log_x = 0

    if iFlag_log_y_in is not None:
        iFlag_log_y = iFlag_log_y_in
    else:
        iFlag_log_y = 0

    if iFlag_scientific_notation_x_in is not None:
        iFlag_scientific_notation_x = iFlag_scientific_notation_x_in
    else:
        iFlag_scientific_notation_x = 0

    if iFlag_scientific_notation_y_in is not None:
        iFlag_scientific_notation_y = iFlag_scientific_notation_y_in
    else:
        iFlag_scientific_notation_y = 0

    if sLabel_x_in is not None:
        sLabel_X = sLabel_x_in
    else:
        sLabel_X = ''

    if sLabel_y_in is not None:
        sLabel_Y = sLabel_y_in
    else:
        sLabel_Y = ''

    if sLabel_legend_in is not None:
        sLabel_legend = sLabel_legend_in
    else:
        sLabel_legend = ''

    if sTitle_in is not None:
        sTitle = sTitle_in
    else:
        sTitle = ''

    fig = plt.figure( dpi=iDPI )
    fig.set_figwidth( iSize_x )
    fig.set_figheight( iSize_y )

    left, width = 0.1, 0.75
    bottom, height = 0.1, 0.75
    spacing = 0.02
    rect_scatter = [left, bottom, width, height]
    rect_histx = [left, bottom + height + spacing, width, 0.15]
    rect_histy = [left + width + spacing, bottom, 0.15, height]


    ax_scatter = plt.axes(rect_scatter)
    ax_scatter.tick_params(direction='in', top=True, right=True)
    ax_histx = plt.axes(rect_histx)
    ax_histx.tick_params(direction='in', labelbottom=False)
    ax_histy = plt.axes(rect_histy)
    ax_histy.tick_params(direction='in', labelleft=False)

    nPoint = len(aData_x)

    x_min = np.nanmin(aData_x)
    x_max = np.nanmax(aData_x)
    y_min = np.nanmin(aData_y)
    y_max = np.ceil(np.nanmax(aData_y))

    x = aData_x
    y = aData_y


    bw="scott"
    gridsize=100
    cut=3
    clip=None

    if clip is None:
        clip = [(-np.inf, np.inf), (-np.inf, np.inf)]
    elif np.ndim(clip) == 1:
        clip = [clip, clip]
    
    aLegend_artist = []
    

    xx, yy, z = scipy_bivariate_kde(x, y , bw, gridsize, cut, clip)
    cmap = plt.get_cmap('BuPu')
    sc = ax_scatter.contourf(xx, yy, z, 5, cmap=cmap) 
    aLegend_artist, aLegend_label = sc.legend_elements(str_format=fmt2)
    aLegend_label_new=[]
    
    
    #aLegend_label_new.append(sLabel)

    ax_scatter.tick_params(axis="x", labelsize=13)
    ax_scatter.tick_params(axis="y", labelsize=13)

    ax_scatter.set_xmargin(0.05)
    ax_scatter.set_ymargin(0.05)

    ax_scatter.set_xlabel(sLabel_X,fontsize=12)
    ax_scatter.set_ylabel(sLabel_Y,fontsize=12)
    ax_scatter.set_title( sTitle, loc='center', fontsize=15)

    # round to nearest years...
    if sFormat_x_in is not None:
        sFormat_x=sFormat_x_in
        ax_scatter.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter(sFormat_x))

  

    if sFormat_y_in is not None:
        #sFormat_y = sFormat_y_in
        #ax_scatter.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter(sFormat_y))
        sFormat_y_dummy =  sFormat_y_in.replace("{", "{x")
        ax_scatter.yaxis.set_major_formatter(mpl.ticker.StrMethodFormatter(  sFormat_y_dummy ) ) 

    
    ax_scatter.tick_params(axis='y', pad=8)
    #ax_scatter.yaxis.set_major_locator(mpl.ticker.MaxNLocator(nbins = 5, prune='lower' ))
    if dMin_x_in is not None:
        dMin_x = dMin_x_in
    else:
        dMin_x = x_min

    if dMax_x_in is not None:
        dMax_x = dMax_x_in
    else:
        dMax_x = x_max

    if dMin_y_in is not None:
        dMin_y = dMin_y_in
    else:
        dMin_y = y_min

    if dMax_y_in is not None:
        dMax_y = dMax_y_in
    else:
        dMax_y = y_max

    ax_scatter.set_xlim( dMin_x, dMax_x )
    ax_scatter.set_ylim( dMin_y, dMax_y)


    if dSpace_x_in is not None:
        dSpace_x = dSpace_x_in
    else:
        dSpace_x = (dMax_x - dMin_x) / 4.0

    if dSpace_y_in is not None:
        dSpace_y = dSpace_y_in
    else:
        dSpace_y = (dMax_y - dMin_y) /4.0

    #ax_scatter.xaxis.set_major_locator(mpl.ticker.MaxNLocator(prune='upper', nbins=5))
    ax_scatter.xaxis.set_major_locator(mpl.ticker.MultipleLocator(base = dSpace_x))

    if iFlag_log_x ==1:
        aLabel_x = []
        for i in np.arange( dMin_x, dMax_x +1, 1 ):
            sTicklabel = r'$10^{{{}}}$'.format(int(i))
            aLabel_x.append(sTicklabel)
            pass

        ax_scatter.set_xticks(np.arange( dMin_x, dMax_x +1, 1 ))
        ax_scatter.set_xticklabels(aLabel_x)
    else:
        if iFlag_scientific_notation_x ==1:
            formatter = mpl.ticker.ScalarFormatter(useMathText=True)
            formatter.set_scientific(True)
            formatter.set_powerlimits((-1,1)) # you might need to change here
            ax_scatter.xaxis.set_major_formatter(formatter)
        else:
            pass
        pass

    ax_scatter.yaxis.set_major_locator(mpl.ticker.MultipleLocator(base = dSpace_y))

    if iFlag_log_y ==1:
        aLabel_y = []
        for i in np.arange( dMin_y, dMax_y +1, 1 ):
            sTicklabel = r'$10^{{{}}}$'.format(int(i))
            aLabel_y.append(sTicklabel)
            pass

        ax_scatter.set_yticks(np.arange( dMin_y, dMax_y +1, 1 ))
        ax_scatter.set_yticklabels(aLabel_y)
        pass
    else:
        if iFlag_scientific_notation_y ==1:
            formatter = mpl.ticker.ScalarFormatter(useMathText=True)
            formatter.set_scientific(True)
            formatter.set_powerlimits((-1,1)) # you might need to change here
            ax_scatter.yaxis.set_major_formatter(formatter)

        pass

    dRatio = (float(iSize_y)/iSize_x) / ( (dMax_y-dMin_y )/ ( dMax_x-dMin_x ) )
    ax_scatter.set_aspect(dRatio, 'box')  #this one set the y / x ratio

    ax_scatter.legend(aLegend_artist, aLegend_label,
                      loc="upper right", fontsize=12)
                      
    slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(x,y)
   
    sText = sLabel_legend
    ax_scatter.text(0.05, 0.95, sText, 
    verticalalignment='top', horizontalalignment='left',
            transform=ax_scatter.transAxes, 
            color='black', fontsize=12)

    sText = r'R: ' + "{:.2f}".format( r_value )
    ax_scatter.text(0.05, 0.9, sText, 
    verticalalignment='top', horizontalalignment='left',
            transform=ax_scatter.transAxes, 
            color='black', fontsize=12)
    
    sText = r'P-value: ' + "{:.2e}".format( p_value )
    ax_scatter.text(0.05, 0.85, sText, 
    verticalalignment='top', horizontalalignment='left',
            transform=ax_scatter.transAxes, 
            color='black', fontsize=12)

    

    ax_scatter.tick_params(which='both', # Options for both major and minor ticks
                           top='off', # turn off top ticks
                           left='off', # turn off left ticks
                           right='off',  # turn off right ticks
                           bottom='off') # turn off bottom ticks
    #ax_histx.set_facecolor('aliceblue')
    density = scipy.stats.gaussian_kde(x)
    xx = np.linspace(dMin_x, dMax_x,1000)
    yy = density(xx)
    ax_histx.plot(xx,yy, color='navy')
    ax_histx.fill_between(xx, yy, 0, linewidth=3,  color = 'lightblue')

    ax_histx.set_xlim( dMin_x, dMax_x )
    ax_histx.set_ylim( 0, auto=None )

    ax_histx.axis('on')
    ax_histx.grid(which='major', color='white', linestyle='-', axis='x')
    ax_histx.xaxis.set_major_locator(mpl.ticker.MultipleLocator(base = dSpace_x/2))
    ax_histx.spines['right'].set_visible(False)
    ax_histx.spines['top'].set_visible(False)
    ax_histx.spines['bottom'].set_visible(False)
    ax_histx.spines['left'].set_visible(False)
    ax_histx.tick_params(which='both', # Options for both major and minor ticks
                         top='off', # turn off top ticks
                         left='off', # turn off left ticks
                         right='off',  # turn off right ticks
                         bottom='off') # turn off bottom ticks
    #ax_histx.axes.get_xaxis().set_visible(False)
    ax_histx.axes.get_yaxis().set_visible(False)
    ax_histx.tick_params(axis='x', colors='white')

    #y margin
    #ax_histy.set_facecolor('aliceblue')
    density = scipy.stats.gaussian_kde(y)
    xx = np.linspace(dMin_y, dMax_y,1000)
    yy = density(xx)
    xx, yy = yy, xx
    ax_histy.plot(xx,yy, color='navy')
    ax_histy.fill_betweenx(yy, 0, xx, linewidth=3,  color = 'lightblue')

    ax_histy.set_xlim(0, auto=None)
    ax_histy.set_ylim(dMin_y, dMax_y)

    ax_histy.axis('on')
    ax_histy.grid(which='major', color='white', linestyle='-', axis='y')
    ax_histy.yaxis.set_major_locator(mpl.ticker.MultipleLocator(base = dSpace_y/2))
    ax_histy.spines['right'].set_visible(False)
    ax_histy.spines['top'].set_visible(False)
    ax_histy.spines['bottom'].set_visible(False)
    ax_histy.spines['left'].set_visible(False)
    ax_histy.axes.get_xaxis().set_visible(False)
    #ax_histy.axes.get_yaxis().set_visible(False)

    ax_histy.tick_params(axis='y', colors='white')

    ax_histy.tick_params(which='both', # Options for both major and minor ticks
                         top='off', # turn off top ticks
                         left='off', # turn off left ticks
                         right='off',  # turn off right ticks
                         bottom='off') # turn off bottom ticks


    plt.savefig(sFilename_out, bbox_inches='tight')

    plt.close('all')
    print('finished plotting')



import numpy as np
from datetime import datetime

from pyearth.toolbox.date.day_in_month import day_in_month

def retrieve_nwis_discharge(sSite, 
                            iYear_start, 
                            iMonth_start, 
                            iDay_start,  
                            iYear_end, 
                            iMonth_end, 
                            iDay_end):
    try:
        import requests
    except ImportError as e:
        raise ImportError(
            "The package 'requests' is required for this function to run.") from e
    
    nwis_dv_url = 'https://waterservices.usgs.gov/nwis/dv/'

    aDate= list()
    sYear_start = "{:04d}".format(iYear_start)
    sMonth_start = "{:02d}".format(iMonth_start)
    sDay_start = "{:02d}".format(iDay_start)
    sYear_end = "{:04d}".format(iYear_end)
    sMonth_end = "{:02d}".format(iMonth_end)
    sDay_end = "{:02d}".format(iDay_end)

    for iYear in range(iYear_start,iYear_end+1):
        for iMonth in range(1,13):
            iDay_end = day_in_month(iYear,iMonth)
            for iDay in range(1,iDay_end+1):            
                aDate.append(datetime(iYear,iMonth,iDay))
    nDay = len(aDate)
    aDate = np.array(aDate)

    start_date = sYear_start + '-' + sMonth_start + '-' + sDay_start
    end_date = sYear_end + '-' + sMonth_end + '-' + sDay_end

    params = {
        'format': 'json',
        'sites': sSite,
        'startDT': start_date,
        'endDT': end_date,
        'parameterCd': '00060' # USGS code for streamflow

    }
    # Send API request
    aDischarge = np.full(nDay, np.nan, dtype= float)
    response = requests.get(nwis_dv_url, params=params)
    unit = None
    # Check if request was successful
    if response.status_code == 200:
        # Parse JSON response
        data = response.json()    
        # Extract streamflow values
        time_series = data['value']['timeSeries'][0]['values'][0]['value']
        unit = data['value']['timeSeries'][0]['variable']['unit']['unitCode']
        print("Unit of measurement:", unit)    
        # Iterate over streamflow values
        for value in time_series:
            date = value['dateTime']
            #take the substring
            date_string = date[0:10]
            # Convert the string to datetime
            iYear = int(date_string[0:4])
            iMonth = int(date_string[5:7])
            iDay = int(date_string[8:10])
            datetime_object = datetime(iYear,iMonth,iDay)
            index = np.where( aDate == datetime_object)
            aDischarge[index] = value['value']  
    else:
        print('Error:', response.status_code)

    return aDischarge, unit


import os, sys
import numpy as np
from osgeo import gdal, osr, ogr, gdalconst

def merge_features(sFilename_in, sFilename_out, sFormat):


    pDataset_in = ogr.Open(sFilename_in)    
    # Get the first layer in the file
    pLayer_in = pDataset_in.GetLayer(0)   
    # Count the number of features (polygons)
    nFeature = pLayer_in.GetFeatureCount()
    # Get the spatial reference of the layer
    pSpatial_reference = pLayer_in.GetSpatialRef()
    wkt2 = pSpatial_reference.ExportToWkt()

    if nFeature == 0:
        print('No feature found in the input shapefile')
        return
    else:
        print('Number of features found in the input shapefile: ' + str(nFeature))

    # Create a new dataset using the output filename
    pDriver = ogr.GetDriverByName(sFormat)
    if pDriver is None:
        print('Driver not found')
        return

    if os.path.exists(sFilename_out):
        pDriver.DeleteDataSource(sFilename_out)

    pDataset_out = pDriver.CreateDataSource(sFilename_out)
    if pDataset_out is None:
        print('Dataset not created')
        return    

    #obtain the geotype 
    iGeomType = pLayer_in.GetGeomType()
    if iGeomType == ogr.wkbPoint:
        #create the layer
        pLayer_out = pDataset_out.CreateLayer('layer', pSpatial_reference, geom_type=ogr.wkbPoint)
        pGeometry_merge = ogr.Geometry(ogr.wkbPoint)
    else:
        if iGeomType == ogr.wkbLineString:
            #create the layer
            pLayer_out = pDataset_out.CreateLayer('layer', pSpatial_reference, geom_type=ogr.wkbLineString)
            pGeometry_merge = ogr.Geometry(ogr.wkbLineString)
        else:
            if iGeomType == ogr.wkbPolygon:
                #create the layer
                pLayer_out = pDataset_out.CreateLayer('layer', pSpatial_reference, geom_type=ogr.wkbPolygon)
                pGeometry_merge = ogr.Geometry(ogr.wkbPolygon)
            else:
                print('Geometry type not supported')
                return
            pass
        pass
    # Create a new layer in the output shapefile
    
    # Loop through the input features and merge them into the output layer
    for i in range(nFeature):
        pFeature = pLayer_in.GetFeature(i)
       
        pGeometry = pFeature.GetGeometryRef()                
        if pGeometry is not None:
          
            # Union the geometry of each feature with the merged polygon
            pGeometry_merge = pGeometry_merge.Union(pGeometry)

    # Create a new feature in the output layer
    pFeature_out = ogr.Feature(pLayer_out.GetLayerDefn())
    pFeature_out.SetGeometry(pGeometry_merge)
    pLayer_out.CreateFeature(pFeature_out)

    #close the dataset
    pDataset_out.Destroy()      
        


    return


import numpy as np
from pyearth.system.define_global_variables import *  

from pyearth.gis.geometry.calculate_distance_based_on_longitude_latitude import calculate_distance_based_on_longitude_latitude



def calculate_spherical_triangle_area(aLongitutde_in, 
                                      aLatitude_in, 
                                      iFlag_radian = None, 
                                      dRadius_in = None):
    """
    Compute the area of a spherical triangle.

    Parameters:
        A, B, C: Tuple or list representing the spherical coordinates (longitude, latitude) of three points.

    Returns:
        The area of the spherical triangle.
    """
    
    if iFlag_radian is None:# convert decimal degrees to radians
        aLongitutde_radian_in = np.radian(aLongitutde_radian_in)
        aLatitude_radian_in = np.radian(aLatitude_in)
    else:
        aLongitutde_radian_in = aLongitutde_in
        aLatitude_radian_in = aLatitude_in
        pass        
    
    dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to = aLongitutde_radian_in[0], aLatitude_radian_in[0], aLongitutde_radian_in[1], aLatitude_radian_in[1]

    a = calculate_distance_based_on_longitude_latitude(dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to, iFlag_radian=1)

    dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to = aLongitutde_radian_in[1], aLatitude_radian_in[1], aLongitutde_radian_in[2], aLatitude_radian_in[2] 

    b = calculate_distance_based_on_longitude_latitude(dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to, iFlag_radian=1)

    dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to = aLongitutde_radian_in[2], aLatitude_radian_in[2], aLongitutde_radian_in[0], aLatitude_radian_in[0] 

    c = calculate_distance_based_on_longitude_latitude(dLongitude_from, dLatitude_from, dLongitude_to, dLatitude_to, iFlag_radian=1)

    s = 0.5 * (a + b + c)

    tanqe = np.sqrt(np.tan(0.5 * s) * np.tan(0.5 * (s - a)) * np.tan(0.5 * (s - b)) * np.tan(0.5 * (s - c)))

    e = 4.0 * np.arctan(tanqe)

    if iFlag_radian is not None: #requested in radian
        return e
    else:       
        if dRadius_in is not None:
            f = e * dRadius_in**2
        else:
            f = e * earth_radius**2
        
        return f

    


import os
from osgeo import osr, ogr

def filter_vector_by_polygon(sFilename_vector_in, sFilename_polygon_in, sFilename_vector_out):
    # Open the input vector layer
    pDriver_parquet = ogr.GetDriverByName('Parquet')
    pDataSource_in = ogr.Open(sFilename_vector_in)
    if pDataSource_in is None:
        print("Could not open input vector")
        return

    pLayer_in = pDataSource_in.GetLayer()

    # Open the clip polygon layer
    pDataSource_clip = ogr.Open(sFilename_polygon_in)
    if pDataSource_clip is None:
        print("Could not open clip polygon")
        pDataSource_in = None
        return

    pLayer_clip = pDataSource_clip.GetLayer()

    # Create a new shapefile for the clipped data
    
    pDataSource_out = pDriver_parquet.CreateDataSource(sFilename_vector_out)
    pLayer_out = pDataSource_out.CreateLayer('clipped', geom_type=ogr.wkbPolygon)

    # Copy the fields from the input layer to the output layer
    pLayer_defn_in = pLayer_in.GetLayerDefn()
    for i in range(pLayer_defn_in.GetFieldCount()):
        field_defn = pLayer_defn_in.GetFieldDefn(i)
        pLayer_out.CreateField(field_defn)   
    
    iFlag_option = 1
    if iFlag_option == 1:
        # Get the extent (bounding box) of the clip layer
        aExtent_clip = pLayer_clip.GetExtent()
        print(aExtent_clip)
        pLayer_in.SetSpatialFilterRect(aExtent_clip[0], aExtent_clip[2], aExtent_clip[1], aExtent_clip[3])
        pass
    else:
        if iFlag_option == 2: #use the outline of the clip layer, this is more accurate
            pass

    # Loop through the input layer and create clipped features
    for feature in pLayer_in:
        pGeometry = feature.GetGeometryRef()        
        pFeature_out = ogr.Feature(pLayer_out.GetLayerDefn())
        pFeature_out.SetGeometry(pGeometry)
        for i in range(pLayer_out.GetLayerDefn().GetFieldCount()):
            pFeature_out.SetField(pLayer_out.GetLayerDefn().GetFieldDefn(i).GetNameRef(),
                                feature.GetField(i))
        pLayer_out.CreateFeature(pFeature_out)
        pFeature_out = None

    # Reset spatial filter and close data sources
    pLayer_in.SetSpatialFilter(None)
    pDataSource_in = pDataSource_clip = pDataSource_out = None


import os
from osgeo import osr, ogr
#accepte user input arguments

def filter_vector_by_attribute(sFilename_input, sFilename_output, sAttribute_name, dValue_filter):
    # Open the input shapefile
    pDriver_parquet = ogr.GetDriverByName('Parquet')
    pDataSource_in = pDriver_parquet.Open(sFilename_input, 0)
    if pDataSource_in is None:
        print("Could not open input file")
        return    
    
    if os.path.exists(sFilename_output):
        os.remove(sFilename_output)
    
    pLayer_in = pDataSource_in.GetLayer()
    #get spatial reference
    pSpatialRef = pLayer_in.GetSpatialRef()
    # Create a new shapefile for the filtered data with the same spatial reference

    pDataSource_out = pDriver_parquet.CreateDataSource(sFilename_output)
    pLayer_out = pDataSource_out.CreateLayer('filtered', geom_type=ogr.wkbPolygon, srs=pSpatialRef)

    # Copy the fields from the input layer to the output layer
    pLayer_defn_in = pLayer_in.GetLayerDefn()
    for i in range(pLayer_defn_in.GetFieldCount()):
        pField_defn = pLayer_defn_in.GetFieldDefn(i)
        pLayer_out.CreateField(pField_defn)

    # Create a SQL query to filter by the attribute value
    sql_query = f"{sAttribute_name} = '{dValue_filter}'"

    # Apply the filter and copy features to the output layer
    pLayer_in.SetAttributeFilter(sql_query)
    for feature in pLayer_in:
        pLayer_out.CreateFeature(feature)

    # Close the data sources
    pDataSource_in = pDataSource_out = None


import os
from osgeo import ogr, gdal
def gdal_read_geojson_boundary(sFilename_boundary_in):
    """
    convert a shpefile to json format.
    This function should be used for stream flowline only.
    """
    iReturn_code = 1
    if os.path.isfile(sFilename_boundary_in):
        pass
    else:
        print('This mesh file does not exist: ', sFilename_boundary_in )
        iReturn_code = 0
        return iReturn_code

    
    pDriver_json = ogr.GetDriverByName('GeoJSON')    
    pDataset_mesh = pDriver_json.Open(sFilename_boundary_in, gdal.GA_ReadOnly)
    pLayer_mesh = pDataset_mesh.GetLayer(0)
    pSpatial_reference_out = pLayer_mesh.GetSpatialRef()
    ldefn = pLayer_mesh.GetLayerDefn()   

    #we also need to spatial reference
    for pFeature_mesh in pLayer_mesh:
        pGeometry_mesh = pFeature_mesh.GetGeometryRef()                     
        pGeometrytype_boundary = pGeometry_mesh.GetGeometryName()
        if(pGeometrytype_boundary == 'POLYGON'):       
            pBoundary_ogr = pGeometry_mesh  
        else:
            if(pGeometrytype_boundary == 'MULTIPOLYGON'):    
                nLine = pGeometry_mesh.GetGeometryCount()
                for i in range(nLine):
                    pBoundary_ogr = pGeometry_mesh.GetGeometryRef(i)
               
                pass
            else:
                pass
            pass   
            
            
    pBoundary_wkt = pBoundary_ogr.ExportToWkt()
    aExtent = pBoundary_ogr.GetEnvelope()
    min_x, max_x, min_y, max_y = aExtent
   
    return pBoundary_wkt, aExtent


import numpy as np

def choose_n_color(nColor_in, pColorMap_in):
    """
    choose color from colormap
    """
    kn = np.arange(nColor_in)

    aColor = [pColorMap_in(float(k)/kn.max()) for k in kn]
    return aColor

def hex_to_RGB(hex):
    ''' "#FFFFFF" -> [255,255,255] '''
    # Pass 16 to the integer function for change of base
    return [int(hex[i:i+2], 16) for i in range(1,6,2)]


def RGB_to_hex(RGB):
    ''' [255,255,255] -> "#FFFFFF" '''
    # Components need to be integers for hex to make sense
    RGB = [int(x) for x in RGB]
    return "#"+"".join(["0{0:x}".format(v) if v < 16 else
                      "{0:x}".format(v) for v in RGB])

def color_dict(gradient):
    ''' Takes in a list of RGB sub-lists and returns dictionary of
    colors in RGB and hex form for use in a graphing function
    defined later on '''
    return {"hex":[RGB_to_hex(RGB) for RGB in gradient],
          "r":[RGB[0] for RGB in gradient],
          "g":[RGB[1] for RGB in gradient],
          "b":[RGB[2] for RGB in gradient]}


def linear_gradient(start_hex, finish_hex="#FFFFFF", n=10):
    ''' returns a gradient list of (n) colors between
    two hex colors. start_hex and finish_hex
    should be the full six-digit color string,
    inlcuding the number sign ("#FFFFFF") '''
    # Starting and ending colors in RGB form
    s = hex_to_RGB(start_hex)
    f = hex_to_RGB(finish_hex)
    # Initilize a list of the output colors with the starting color
    RGB_list = [s]
    # Calcuate a color at each evenly spaced value of t from 1 to n
    for t in range(1, n):
        # Interpolate RGB vector for color at the current value of t
        curr_vector = [  int(s[j] + (float(t)/(n-1))*(f[j]-s[j]))     for j in range(3)   ]
        # Add it to our list of output colors
        RGB_list.append(curr_vector)

    return color_dict(RGB_list)

def rand_hex_color(num=1):
    ''' Generate random hex colors, default is one,
      returning a string. If num is greater than
      1, an array of strings is returned. '''
    colors = [   RGB_to_hex([x*255 for x in np.random.rand(3)])    for i in range(num)  ]
    if num == 1:
       return colors[0]
    else:
       return colors

def polylinear_gradient(colors, n):
    ''' returns a list of colors forming linear gradients between
      all sequential pairs of colors. "n" specifies the total
      number of desired output colors '''
    # The number of colors per individual linear gradient
    n_out = int(float(n) / (len(colors) - 1))
    # returns dictionary defined by color_dict()
    gradient_dict = linear_gradient(colors[0], colors[1], n_out)

    if len(colors) > 1:
        for col in range(1, len(colors) - 1):
            next = linear_gradient(colors[col], colors[col+1], n_out)
            for k in ("hex", "r", "g", "b"):
            # Exclude first point to avoid duplicates
                gradient_dict[k] += next[k][1:]

    return gradient_dict





